﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Linq;
using System.ServiceProcess;
using System.Text;
using System.Threading.Tasks;
using System.Xml.Serialization;
using System.IO;
using Bsc;
using System.Timers;
using System.Net.NetworkInformation;
using System.Net;
using System.Net.Mail;
using System.Management;
using System.Collections.Specialized;
using System.Collections;
using System.Threading;
using System.Reflection;
using OpenPop.Pop3;
using System.Text.RegularExpressions;
using System.Data.SqlClient;
using System.Security;

namespace tms
{
    public partial class ServiceMain : ServiceBase
    {
        tmsParams myParams;
        EventLog eventLogMain = new EventLog();
        SqlConnection mssql = null;


        public class tmsState
        {
            public enum timerTypeEnum { normalTimer, controlTimer, listenerTimer };
            public enum mailCommandTypeEnum { mcListAllStates };
            public oneNode nodeParams;
            public Int64 pingFailed;
            public Int64 jobsFailed;
            public Dictionary<string, Int64> serviceFailed;
            public Int64 spaceFailed;
            public Int64 csFailed;
            public Int64 cspFailed;
            public Int64 cpjFailed;
            public bool busy;
            public timerTypeEnum timerType;
            public DateTime lastActivity;
        }

        Dictionary<System.Timers.Timer, tmsState> timerPool;
        System.Timers.Timer timerControl;
        System.Timers.Timer timerListener;

        public ServiceMain()
        {
            InitializeComponent();

        }


        public SqlConnection connectSQL()
        {
            bool error = false;

            SecureString secureString = new SecureString();
            foreach (char c in myParams.SQLPassword.ToCharArray())
            {
                secureString.AppendChar(c);
            }

            secureString.MakeReadOnly();

            SqlConnection con = null;

            try
            {
                con = new SqlConnection(myParams.SQLconnection, new SqlCredential(myParams.SQLUser, secureString));
            }
            catch
            {
                error = true;
            }


            if ((!error) && (con != null))
            {

                try
                {
                    con.Open();
                }
                catch
                {
                    error = true;
                }

            }
            else return null;


            if ((!error) && (con.State == ConnectionState.Open))
            {
                return con;
            }
            else return null;


        }


        public void disconnectSQL(SqlConnection con)
        {

            try
            {
                if (con != null)
                {
                    con.Close();

                }
            }
            catch
            { };


        }

        public bool writeSQL(SqlConnection con, DateTime events_dt, string events_object, string events_problem, string events_pdesc, Int64 events_pcount, bool events_critical, bool events_ok)
        {
            bool error = false;

            if (con != null)
            {

                try
                {
                    SqlCommand command = new SqlCommand("insert into events (events_dt, events_object, events_problem, events_pdesc, events_pcount, events_critical, events_ok) values (@events_dt, @events_object, @events_problem, @events_pdesc, @events_pcount, @events_critical, @events_ok)", con);

                    command.Parameters.AddWithValue("@events_dt", events_dt);
                    command.Parameters.AddWithValue("@events_object", events_object);
                    command.Parameters.AddWithValue("@events_problem", events_problem);
                    command.Parameters.AddWithValue("@events_pdesc", events_pdesc);
                    command.Parameters.AddWithValue("@events_pcount", events_pcount);
                    command.Parameters.AddWithValue("@events_critical", events_critical);
                    command.Parameters.AddWithValue("@events_ok", events_ok);
                    command.ExecuteNonQuery();
                }
                catch
                {
                    error = true;
                }

            }
            else error = true;
            return error;
        }

        public void doMailCommand(tmsState.mailCommandTypeEnum mc, string addr)
        {

            string s = "";


            s = "<html><body>";



            if (timerPool != null)
            {

                try
                {

                    if (mc == tmsState.mailCommandTypeEnum.mcListAllStates)
                    {

                        s += "<table border=1 cellpadding=2 cellspacing=1>";

                        foreach (System.Timers.Timer t in timerPool.Keys)
                        {
                            tmsState st = timerPool[t];



                            if ((st.timerType == tmsState.timerTypeEnum.normalTimer) && (st.nodeParams.isActive))
                            {


                                string connectBy = st.nodeParams.useIP ? st.nodeParams.ip : st.nodeParams.dnsname;


                                s += string.Format("<tr><td colspan=2 bgcolor=\"yellow\" align=\"center\"><strong>{0}</strong></td></tr>", connectBy);
                                s += string.Format("<tr><td>Last activity</td><td><font color=\"{1}\">{0}</td></tr>", st.lastActivity, (DateTime.Now - st.lastActivity).Minutes > myParams.timerInterval * 2 ? "red" : "green");
                                s += string.Format("<tr><td>DNS name</td><td>{0}</td></tr>", st.nodeParams.dnsname);
                                s += string.Format("<tr><td>IP</td><td>{0}</td></tr>", st.nodeParams.ip);
                                s += string.Format("<tr><td>Description</td><td>{0}</td></tr>", st.nodeParams.description);
                                s += string.Format("<tr><td>Busy now</td><td>{0}</td></tr>", st.busy);

                                if (st.nodeParams.controlPing)
                                    s += string.Format("<tr><td>Ping failed</td><td bgcolor=\"{1}\">{0}</td></tr>", st.pingFailed, st.pingFailed != 0 ? "#FFCCCC" : "#99FF99");

                                if (st.nodeParams.controlSpace)
                                {
                                    s += string.Format("<tr><td>Space failed</td><td bgcolor=\"{1}\">{0}</td></tr>", st.spaceFailed, st.spaceFailed != 0 ? "#FFCCCC" : "#99FF99");
                                    s += string.Format("<tr><td>Space get info failed</td><td bgcolor=\"{1}\">{0}</td></tr>", st.cspFailed, st.cspFailed != 0 ? "#FFCCCC" : "#99FF99");
                                }

                                if (st.nodeParams.controlJobs)
                                {
                                    s += string.Format("<tr><td>Jobs failed</td><td bgcolor=\"{1}\">{0}</td></tr>", st.jobsFailed, st.jobsFailed != 0 ? "#FFCCCC" : "#99FF99");
                                    s += string.Format("<tr><td>Jobs get info failed</td><td bgcolor=\"{1}\">{0}</td></tr>", st.cpjFailed, st.cpjFailed != 0 ? "#FFCCCC" : "#99FF99");
                                }

                                if (st.nodeParams.controlService)
                                {

                                    s += string.Format("<tr><td colspan=2 align=\"center\"><strong>Services</strong></td></tr>");

                                    s += string.Format("<tr><td>Service get info failed</td><td bgcolor=\"{1}\">{0}</td></tr>", st.csFailed, st.csFailed != 0 ? "#FFCCCC" : "#99FF99");

                                    foreach (string srv in st.serviceFailed.Keys)
                                    {
                                        s += string.Format("<tr><td>{0}</td><td bgcolor=\"{2}\">{1}</td></tr>", srv, st.serviceFailed[srv], st.serviceFailed[srv] != 0 ? "#FFCCCC" : "#99FF99");
                                    }


                                }

                                s += "<tr><td colspan=2>&nbsp;</td></tr>";


                            }



                        }

                        s += "</table>";

                        s += "</body><html>";

                        sendMailDirect("TMS report", s, null, addr);

                        if (myParams.debugMessages)
                            addEvent(String.Format("Mail command report send to {0}", addr), false);

                    }

                }

                catch (Exception exc)
                {


                    addEvent(String.Format("Check free space proxy exception {0} with source {1}, ts = {2}, st = {3}, path = {4}", exc.Message, exc.Source, exc.TargetSite, exc.StackTrace, addr), true);
                }

            }



        }

        protected override void OnStart(string[] args)
        {
            addEvent("Service started at " + DateTime.Now.ToString("HH:mm:ss dd.MM.yyyy"));

            string[] rp = { "-config" };
            CommandLineArgumentParser.DefineRequiredParameters(rp);
            var myArgs = Environment.GetCommandLineArgs().ToList();
            myArgs.RemoveAt(0);
            CommandLineArgumentParser.ParseArguments(myArgs.ToArray());

            var configPath = CommandLineArgumentParser.GetParamValue("-config");
            addEvent("Using config " + configPath);

            XmlSerializer serializer = new XmlSerializer(typeof(tmsParams));
            FileStream fs = new FileStream(configPath, FileMode.Open);
            myParams = (tmsParams)serializer.Deserialize(fs);
            fs.Close();

            if (myParams.debugMessages)
                addEvent(String.Format("Found {0} nodes", myParams.nodes.Count));

            timerPool = new Dictionary<System.Timers.Timer, tmsState>();

            foreach (oneNode thisNode in myParams.nodes)
            {
                if (thisNode.isActive)
                {

                    var timer1 = new System.Timers.Timer();
                    timer1.Enabled = false;

                    timer1.Interval = TimeSpan.FromMinutes(myParams.timerInterval).TotalMilliseconds;
                    timer1.Elapsed += new System.Timers.ElapsedEventHandler(timer_Elapsed);
                    timer1.AutoReset = false;

                    var state1 = new tmsState();
                    state1.timerType = tmsState.timerTypeEnum.normalTimer;
                    state1.nodeParams = thisNode;
                    state1.pingFailed = 0;
                    state1.jobsFailed = 0;
                    state1.spaceFailed = 0;
                    state1.csFailed = 0;
                    state1.cspFailed = 0;
                    state1.cpjFailed = 0;

                    state1.serviceFailed = new Dictionary<string, Int64>(StringComparer.CurrentCultureIgnoreCase);

                    foreach (string sn in thisNode.serviceList)
                        state1.serviceFailed.Add(sn, 0);

                    state1.busy = false;
                    state1.lastActivity = DateTime.Now;
                    timerPool.Add(timer1, state1);
                }
            }


            foreach (System.Timers.Timer timer in timerPool.Keys)
            {
                timer.Enabled = true;

            }


            timerControl = new System.Timers.Timer();
            timerControl.Enabled = false;

            timerControl.Interval = TimeSpan.FromMinutes(myParams.timerControlInterval).TotalMilliseconds;
            timerControl.Elapsed += new System.Timers.ElapsedEventHandler(timerControl_Elapsed);
            timerControl.AutoReset = false;

            var stateControl = new tmsState();
            stateControl.timerType = tmsState.timerTypeEnum.controlTimer;
            stateControl.nodeParams = null;
            stateControl.pingFailed = 0;
            stateControl.jobsFailed = 0;

            stateControl.serviceFailed = null;

            stateControl.busy = false;
            timerPool.Add(timerControl, stateControl);
            timerControl.Start();


            timerListener = new System.Timers.Timer();
            timerListener.Enabled = false;

            timerListener.Interval = TimeSpan.FromMinutes(myParams.popListenerTimerInterval).TotalMilliseconds;
            timerListener.Elapsed += new System.Timers.ElapsedEventHandler(timerListener_Elapsed);
            timerListener.AutoReset = false;

            var stateListener = new tmsState();
            stateListener.timerType = tmsState.timerTypeEnum.listenerTimer;
            stateListener.nodeParams = null;
            stateListener.pingFailed = 0;
            stateListener.jobsFailed = 0;

            stateListener.serviceFailed = null;

            stateListener.busy = false;
            timerPool.Add(timerListener, stateListener);
            timerListener.Start();


            if (myParams.writeToSQL)
            {

                mssql = connectSQL();

                if (mssql == null)
                {
                    addEvent("Error when trying to connect with SQL server", EventLogEntryType.Warning);
                }
                else
                {
                    writeSQL(mssql, DateTime.Now, "TMS", "Startup", "Monitor started", 0, false, true);

                }
            }


        }

        private void timerListener_Elapsed(object sender, ElapsedEventArgs e)
        {
            if (timerPool != null)
            {

                tmsState cs = timerPool[sender as System.Timers.Timer];

                if (!cs.busy)
                {

                    cs.busy = true;

                    checkIncommingEmail();

                    cs.busy = false;

                    cs.lastActivity = DateTime.Now;
                }

            }

            (sender as System.Timers.Timer).Start();


        }

        protected override void OnStop()
        {

            if (timerPool != null)
            {

                foreach (System.Timers.Timer timer in timerPool.Keys)
                {
                    timer.Enabled = false;

                }
            }

            addEvent("Service stopped at " + DateTime.Now.ToString("HH:mm:ss dd.MM.yyyy"));

            if (myParams.writeToSQL)
            {
                writeSQL(mssql, DateTime.Now, "TMS", "Shutdown", "Monitor stopped", 0, false, true);
                disconnectSQL(mssql);

            }

        }


        public void addEvent(string log, bool error = false, EventLogEntryType type = EventLogEntryType.Information)
        {

            try
            {

                if (!EventLog.SourceExists("TMSrvSource"))
                {
                    EventLog.CreateEventSource("TMSrvSource", "TMSrvLog");
                    this.Stop();
                }

                eventLogMain.Source = "TMSrvSource";

                if (!error)
                    eventLogMain.WriteEntry(log, type);
                else eventLogMain.WriteEntry(log, EventLogEntryType.Error);

            }
            catch
            {


            }



        }

        public void addEvent(string log, EventLogEntryType type)
        {

            addEvent(log, false, type);

        }



        private bool checkFreeSpaceProxy(string path, Dictionary<string, double> drives, Int32 spaceLimit, TimeSpan wmiTimeOut, TimeSpan waitTime, out bool withError)
        {

            bool error = false;

            try
            {

                Stopwatch sw = new Stopwatch();


                Int32 csInterval = Convert.ToInt32(waitTime.TotalMilliseconds);
                bool csFlag = false;
                Int32 tryNo = 0;

                sw.Start();

                while ((sw.ElapsedMilliseconds <= csInterval) && (!csFlag))
                {

                    tryNo++;

                    if (myParams.debugMessages)
                        addEvent(string.Format("Trying to check free space on path {0}, try No. {1}", path, tryNo));

                    drives.Clear();

                    csFlag = checkFreeSpace(path, drives, spaceLimit, wmiTimeOut, out error);

                }

                sw.Stop();

                withError = error;
                return csFlag;

            }

            catch (Exception exc)
            {
                addEvent(String.Format("Check free space proxy exception {0} with source {1}, ts = {2}, st = {3}, path = {4}", exc.Message, exc.Source, exc.TargetSite, exc.StackTrace, path), true);

                withError = true;
                return false;

            }



        }


        private bool checkServicesProxy(string path, List<string> names, Dictionary<string, bool> statuses, TimeSpan wmiTimeOut, TimeSpan waitTime)
        {
            try
            {

                Stopwatch sw = new Stopwatch();


                Int32 csInterval = Convert.ToInt32(waitTime.TotalMilliseconds);
                bool csFlag = false;
                Int32 tryNo = 0;

                sw.Start();

                while ((sw.ElapsedMilliseconds <= csInterval) && (!csFlag))
                {

                    tryNo++;

                    if (myParams.debugMessages)
                        addEvent(string.Format("Trying to check services on path {0}, try No. {1}", path, tryNo));

                    statuses.Clear();

                    foreach (string s in names)
                    {
                        statuses.Add(s, false);
                    };

                    csFlag = checkServices(path, names.ToArray(), statuses, wmiTimeOut);

                }

                sw.Stop();

                return csFlag;

            }

            catch (Exception exc)
            {
                addEvent(String.Format("Check services proxy exception {0} with source {1}, ts = {2}, st = {3}, path = {4}", exc.Message, exc.Source, exc.TargetSite, exc.StackTrace, path), true);
                return false;

            }

        }


        private bool checkServices(string path, string[] names, Dictionary<string, bool> statuses, TimeSpan wmiTimeOut)
        {

            bool res = true;

            try
            {

                var resetEvent = new ManualResetEvent(false);

                ThreadPool.QueueUserWorkItem(
        arg =>
        {

            try
            {

                string providerPath = String.Format(@"\\{0}\root\CIMv2", path);

                string query = "select Name, State from Win32_Service Where";
                for (int i = 0; i != names.Count(); i++)
                {
                    if (i == 0) query += string.Format(" Name = '{0}'", names[i]);
                    else query += string.Format(" OR Name = '{0}'", names[i]);

                }

                ConnectionOptions options = new ConnectionOptions();
                // options.Timeout = new TimeSpan(0, 1, 0);

                ManagementScope scope = new ManagementScope(providerPath, options);

                ManagementObjectSearcher searcher =
           new ManagementObjectSearcher(scope, new ObjectQuery(query));
                searcher.Options.Timeout = wmiTimeOut;
                searcher.Options.ReturnImmediately = false;


                ManagementObjectCollection retObjectCollection = searcher.Get();
                foreach (ManagementObject mo in retObjectCollection)
                {

                    statuses[mo.GetPropertyValue("Name") as String] = (mo.GetPropertyValue("State") as String == "Running");
                }

            }

            catch (Exception exc)
            {
                addEvent(String.Format("Check services exception {0} with source {1}, ts = {2}, st = {3}, path = {4}", exc.Message, exc.Source, exc.TargetSite, exc.StackTrace, path), true);
                res = false;

            };

            resetEvent.Set();


        });

                resetEvent.WaitOne();


            }
            catch (Exception exc)
            {
                addEvent(String.Format("Check services exception {0} with source {1}, ts = {2}, st = {3}, path = {4}", exc.Message, exc.Source, exc.TargetSite, exc.StackTrace, path), true);
                res = false;

            }

            return res;

        }



        private bool checkFreeSpace(string path, Dictionary<string, double> drives, Int32 spaceLimit, TimeSpan wmiTimeOut, out bool withError)
        {

            bool res = true;
            bool error = false;


            try
            {

                var resetEvent = new ManualResetEvent(false);


                ThreadPool.QueueUserWorkItem(
        arg =>
        {

            try
            {


                string providerPath = String.Format(@"\\{0}\root\CIMv2", path);

                string query = "Select Name, Size, FreeSpace from Win32_LogicalDisk Where DriveType = 3";


                ConnectionOptions options = new ConnectionOptions();
                //options.Timeout = new TimeSpan(0, 1, 0);

                ManagementScope scope = new ManagementScope(providerPath, options);


                ManagementObjectSearcher searcher =
           new ManagementObjectSearcher(scope, new ObjectQuery(query));

                searcher.Options.Timeout = options.Timeout = wmiTimeOut;
                searcher.Options.ReturnImmediately = false;

                ManagementObjectCollection retObjectCollection = searcher.Get();
                foreach (ManagementObject mo in retObjectCollection)
                {
                    UInt64 tfs = (UInt64)mo.GetPropertyValue("FreeSpace");
                    UInt64 ts = (UInt64)mo.GetPropertyValue("Size");

                    double driveFree = ((double)tfs / (double)ts * 100);

                    if (driveFree < spaceLimit) res = false;

                    drives.Add(mo.GetPropertyValue("Name") as String, driveFree);


                }



            }

            catch (Exception exc)
            {
                addEvent(String.Format("Check free space exception {0} with source {1}, ts = {2}, st = {3}, path = {4}", exc.Message, exc.Source, exc.TargetSite, exc.StackTrace, path), true);
                res = false;
                error = true;
            }

            resetEvent.Set();

        });

                resetEvent.WaitOne();


            }

            catch (Exception exc)
            {
                addEvent(String.Format("Check free space exception {0} with source {1}, ts = {2}, st = {3}, path = {4}", exc.Message, exc.Source, exc.TargetSite, exc.StackTrace, path), true);
                res = false;
                error = true;

            }

            withError = error;
            return res;

        }


        public double checkPing(string addr)
        {




            int cntOk = 0;
            int cntTotal = 0;

            try
            {


                var resetEvent = new ManualResetEvent(false);


                ThreadPool.QueueUserWorkItem(
        arg =>
        {

            var pinger = new Ping();

            Stopwatch sw = new Stopwatch();
            sw.Start();

            Int32 pingInterval = Convert.ToInt32(TimeSpan.FromSeconds(myParams.pingInterval).TotalMilliseconds);

            while (sw.ElapsedMilliseconds <= pingInterval)
            {
                try
                {
                    PingReply pr = pinger.Send(addr);
                    if (pr.Status == IPStatus.Success) cntOk++;
                    cntTotal++;


                }

                // catch (PingException pe)
                catch
                {
                    if (cntTotal == 0) cntTotal = 1;
                    break;
                }

                Thread.Sleep(1);

            }

            sw.Stop();


            resetEvent.Set();

        });

                resetEvent.WaitOne();

            }
            catch (Exception exc)
            {
                addEvent(String.Format("Check ping exception {0} with source {1}, ts = {2}, st = {3}, path = {4}", exc.Message, exc.Source, exc.TargetSite, exc.StackTrace, addr), true);
                return (double)0;

            }

            return ((double)cntOk / (double)cntTotal * 100);

        }


        private void timerControl_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        {

            if (timerPool != null)
            {

                tmsState cs = timerPool[sender as System.Timers.Timer];

                if (!cs.busy)
                {

                    cs.busy = true;

                    foreach (System.Timers.Timer tp in timerPool.Keys)
                    {

                        tmsState st = timerPool[tp];

                        if ((st.timerType == tmsState.timerTypeEnum.normalTimer) && (st.nodeParams.isActive == true))
                        {

                            string connectBy;

                            if (st.nodeParams.useIP)
                                connectBy = st.nodeParams.ip;
                            else connectBy = st.nodeParams.dnsname;


                            TimeSpan diff = DateTime.Now - st.lastActivity;
                            if (diff.Minutes > myParams.timerInterval * 2)
                            {
                                addEvent(string.Format("Timer died for {0}, last run {1}", connectBy, st.lastActivity.ToString("HH:mm:ss dd.MM.yyyy")), true);

                            }
                            else
                            {
                                if (myParams.debugMessages)
                                    addEvent(string.Format("Timer ok for {0}, last run {1}", connectBy, st.lastActivity.ToString("HH:mm:ss dd.MM.yyyy")));

                            }


                        }


                        if ((st.timerType == tmsState.timerTypeEnum.listenerTimer) && myParams.enablePopListener)
                        {

                            TimeSpan diff = DateTime.Now - st.lastActivity;
                            if (diff.Minutes > myParams.popListenerTimerInterval * 2)
                            {
                                addEvent(string.Format("Timer listener frozen, last run {0}", st.lastActivity.ToString("HH:mm:ss dd.MM.yyyy")), true);

                            }
                            else
                            {
                                if (myParams.debugMessages)
                                    addEvent(string.Format("Timer listener ok, last run {0}", st.lastActivity.ToString("HH:mm:ss dd.MM.yyyy")));

                            }


                        }



                    }

                    cs.busy = false;
                }

            }


            if (myParams.writeToSQL)
            {

                if (mssql == null) addEvent("SQL connection is null", true);
                else
                {
                    if (mssql.State == ConnectionState.Broken || mssql.State == ConnectionState.Closed)
                    {
                        addEvent("SQL connection is broken or closed", true);
                    }

                }


            }


            (sender as System.Timers.Timer).Start();


        }



        private void timer_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        {

            if (timerPool != null)
            {

                tmsState ts = timerPool[sender as System.Timers.Timer];


                ts.lastActivity = DateTime.Now;

                string connectBy;

                if (ts.nodeParams.useIP)
                    connectBy = ts.nodeParams.ip;
                else connectBy = ts.nodeParams.dnsname;

                if (!ts.busy)
                {


                    try
                    {


                        ts.busy = true;

                        if (myParams.debugMessages)
                            addEvent("Timer elapsed at " + DateTime.Now.ToString("HH:mm:ss dd.MM.yyyy") + " for " + connectBy + ", checking");


                        if (ts.nodeParams.controlPing)
                        {


                            double pingStat = checkPing(connectBy);

                            if (pingStat < ts.nodeParams.pingLimit)
                            {

                                ts.pingFailed++;

                                addEvent(String.Format("Ping problem at {0}, percent is {1:0.##}, repeat No. {2}", connectBy, pingStat, ts.pingFailed), true);

                                if (myParams.writeToSQL)
                                    writeSQL(mssql, DateTime.Now, connectBy, "Ping", String.Format("Ping problem at {0}, percent is {1:0.##}, repeat No. {2}", connectBy, pingStat, ts.pingFailed), ts.pingFailed, true, false);


                                if (ts.nodeParams.doEmail && ts.pingFailed <= myParams.pingReports)
                                    sendMailFormat(string.Format("Ping failed at {0}", connectBy), connectBy, ts.nodeParams.description, DateTime.Now, "Ping", pingStat == 0 ? "The system does not respond to ping" : String.Format("Ping statisctics is bad: {0:0.##} % has reached", pingStat), ts.pingFailed, true);

                            }
                            else
                            {
                                if (ts.pingFailed != 0)
                                {
                                    addEvent(String.Format("Ping NOW OK after FAIL at {0}, percent is {1:0.##}", connectBy, pingStat), EventLogEntryType.Warning);

                                    if (myParams.writeToSQL)
                                        writeSQL(mssql, DateTime.Now, connectBy, "Ping", String.Format("Ping NOW OK after FAIL at {0}, percent is {1:0.##}", connectBy, pingStat), ts.pingFailed, false, true);

                                    if (myParams.sendOK)
                                        sendMailFormat(string.Format("Ping now OK at {0}", connectBy), connectBy, ts.nodeParams.description, DateTime.Now, "Ping", String.Format("Ping statisctics is good: {0:0.##} %", pingStat), ts.pingFailed, false);
                                }


                                else
                                {
                                    if (myParams.debugMessages)
                                        addEvent(String.Format("Ping OK at {0}, percent is {1:0.##}", connectBy, pingStat));
                                }
                                ts.pingFailed = 0;
                            }



                        }

                        bool doOtherChecks = (myParams.forceOnBadPing) || ((!ts.nodeParams.controlPing) || (ts.pingFailed == 0));

                        if (!doOtherChecks)
                        {
                            if (myParams.debugMessages)
                                addEvent(String.Format("Only ping applied to {0}, other checks are skipped", connectBy));
                        }

                        if (ts.nodeParams.controlService && doOtherChecks)
                        {
                            Dictionary<string, bool> servRes = new Dictionary<string, bool>(StringComparer.CurrentCultureIgnoreCase);


                            bool csflag = checkServicesProxy(connectBy, ts.nodeParams.serviceList, servRes, new TimeSpan(ts.nodeParams.wmiTimeOut * TimeSpan.TicksPerSecond), new TimeSpan(TimeSpan.TicksPerSecond * myParams.serviceTryTime));

                            if (!csflag) ts.csFailed++; else ts.csFailed = 0;

                            if (csflag)
                            {

                                foreach (string serviceName in servRes.Keys)
                                {

                                    if (!(bool)servRes[serviceName])
                                    {
                                        ts.serviceFailed[serviceName]++;


                                        addEvent(String.Format("Service problem at {0}, name is {1}, repeat No. {2}", connectBy, serviceName, ts.serviceFailed[serviceName]), true);




                                        if (myParams.writeToSQL)
                                            writeSQL(mssql, DateTime.Now, connectBy, "Service", String.Format("Service problem at {0}, name is {1}, repeat No. {2}", connectBy, serviceName, ts.serviceFailed[serviceName]), ts.serviceFailed[serviceName], true, false);

                                        if ((ts.nodeParams.doEmail) && (ts.serviceFailed[serviceName] <= myParams.serviceReports) && ((ts.pingFailed == 0) | (!ts.nodeParams.controlPing)))
                                            sendMailFormat(string.Format("Service {0} failed at {1}", serviceName, connectBy), connectBy, ts.nodeParams.description, DateTime.Now, "Service", String.Format("Service stopped or no information for {0}", serviceName), ts.serviceFailed[serviceName], true);




                                    }
                                    else
                                    {
                                        if (ts.serviceFailed[serviceName] != 0)
                                        {
                                            addEvent(String.Format("Service NOW OK after FAIL at {0}, name is {1}", connectBy, serviceName), EventLogEntryType.Warning);

                                            if (myParams.writeToSQL)
                                                writeSQL(mssql, DateTime.Now, connectBy, "Service", String.Format("Service NOW OK after FAIL at {0}, name is {1}", connectBy, serviceName), ts.serviceFailed[serviceName], false, true);

                                            if (myParams.sendOK)
                                                sendMailFormat(string.Format("Service {0} now OK at {1}", serviceName, connectBy), connectBy, ts.nodeParams.description, DateTime.Now, "Service", String.Format("Service is running {0}", serviceName), ts.serviceFailed[serviceName], false);
                                        }
                                        else
                                        {
                                            if (myParams.debugMessages)
                                                addEvent(String.Format("Service OK at {0}, name is {1}", connectBy, serviceName));
                                        }
                                        ts.serviceFailed[serviceName] = 0;

                                    }


                                }

                            }
                            else
                            {

                                if (myParams.writeToSQL)
                                    writeSQL(mssql, DateTime.Now, connectBy, "Service", String.Format("Service get info failed at {0}", connectBy), ts.csFailed, false, false);

                                if (myParams.reportWhenNoInfo)
                                {

                                    if ((ts.nodeParams.doEmail) && (ts.csFailed <= myParams.serviceReports) && ((ts.pingFailed == 0) | (!ts.nodeParams.controlPing)))
                                        sendMailFormat(string.Format("Service get info failed at {0}", connectBy), connectBy, ts.nodeParams.description, DateTime.Now, "Service", String.Format("No information for services"), ts.csFailed, true);
                                }
                            }



                        }



                        if (ts.nodeParams.controlSpace && doOtherChecks)
                        {


                            Dictionary<string, double> drives = new Dictionary<string, double>(StringComparer.CurrentCultureIgnoreCase);



                            bool csperror = false;
                            bool freeSpace = checkFreeSpaceProxy(connectBy, drives, ts.nodeParams.spaceLimit, new TimeSpan(ts.nodeParams.wmiTimeOut * TimeSpan.TicksPerSecond), new TimeSpan(TimeSpan.TicksPerSecond * myParams.spaceTryTime), out csperror);


                            if (csperror) ts.cspFailed++; else ts.cspFailed = 0;

                            string spaces = "";

                            foreach (string s in drives.Keys)
                            {
                                if (spaces == "")
                                    spaces += string.Format("{0} {1:0.##}%", s, drives[s]);
                                else
                                    spaces += string.Format(", {0} {1:0.##}%", s, drives[s]);
                            }



                            if (!csperror)
                            {

                                if (!freeSpace)
                                {
                                    ts.spaceFailed++;


                                    addEvent(String.Format("Space problem at {0}, percents are {1}, repeat No. {2}", connectBy, spaces == "" ? "unknown" : spaces, ts.spaceFailed), true);


                                    if (myParams.writeToSQL)
                                        writeSQL(mssql, DateTime.Now, connectBy, "Space", String.Format("Space problem at {0}, percents are {1}, repeat No. {2}", connectBy, spaces == "" ? "unknown" : spaces, ts.spaceFailed), ts.spaceFailed, true, false);


                                    if ((ts.nodeParams.doEmail && ts.spaceFailed <= myParams.spaceReports) && (ts.pingFailed == 0) && ((ts.pingFailed == 0) | (!ts.nodeParams.controlPing)))
                                        sendMailFormat(string.Format("Space failed at {0}", connectBy), connectBy, ts.nodeParams.description, DateTime.Now, "Disk space", String.Format("Disk space too low: {0}", spaces == "" ? "unknown" : spaces), ts.spaceFailed, true);



                                }
                                else
                                {
                                    if (ts.spaceFailed != 0)
                                    {
                                        addEvent(String.Format("Space NOW OK after FAIL at {0}, percents are {1}", connectBy, spaces), EventLogEntryType.Warning);

                                        if (myParams.writeToSQL)
                                            writeSQL(mssql, DateTime.Now, connectBy, "Space", String.Format("Space NOW OK after FAIL at {0}, percents are {1}", connectBy, spaces), ts.spaceFailed, false, true);

                                        if (myParams.sendOK)
                                            sendMailFormat(string.Format("Space now OK at {0}", connectBy), connectBy, ts.nodeParams.description, DateTime.Now, "Disk space", String.Format("Disk space enough: {0}", spaces), ts.spaceFailed, false);
                                    }
                                    else
                                    {
                                        if (myParams.debugMessages)
                                            addEvent(String.Format("Space OK at {0}, percents are {1}", connectBy, spaces));
                                    }
                                    ts.spaceFailed = 0;

                                }

                            }
                            else
                            {

                                addEvent(String.Format("Space get info failed at {0}", connectBy), EventLogEntryType.Warning);

                                if (myParams.writeToSQL)
                                    writeSQL(mssql, DateTime.Now, connectBy, "Space", String.Format("Space get info failed at {0}", connectBy), ts.cspFailed, false, false);

                                if (myParams.reportWhenNoInfo)
                                {

                                    if ((ts.nodeParams.doEmail) && (ts.cspFailed <= myParams.spaceReports) && ((ts.pingFailed == 0) | (!ts.nodeParams.controlPing)))
                                        sendMailFormat(string.Format("Space get info failed at {0}", connectBy), connectBy, ts.nodeParams.description, DateTime.Now, "Space", String.Format("No information for disk space"), ts.cspFailed, true);
                                }

                            }


                        }




                        if (ts.nodeParams.controlJobs && doOtherChecks)
                        {


                            Dictionary<string, Int32> frozen = new Dictionary<string, int>(StringComparer.CurrentCultureIgnoreCase);

                            bool cpjerror = false;
                            bool jobsOk = checkJobsProxy(connectBy, ts.nodeParams.jobList.ToArray(), frozen, new TimeSpan(ts.nodeParams.wmiTimeOut * TimeSpan.TicksPerSecond), new TimeSpan(myParams.jobsTryTime * TimeSpan.TicksPerSecond), ts.nodeParams.jobLimit, out cpjerror);


                            if (cpjerror) ts.cpjFailed++; else ts.cpjFailed = 0;

                            string jobs = "";

                            foreach (string s in frozen.Keys)
                            {
                                if (jobs == "")
                                    jobs += string.Format("{0} - {1} min", s, frozen[s]);
                                else
                                    jobs += string.Format(", {0} - {1} min", s, frozen[s]);
                            }



                            if (!cpjerror)
                            {

                                if (!jobsOk)
                                {
                                    ts.jobsFailed++;


                                    addEvent(String.Format("Print jobs problem at {0}, times are {1}, repeat No. {2}", connectBy, jobs == "" ? "unknown" : jobs, ts.jobsFailed), true);

                                    if (myParams.writeToSQL)
                                        writeSQL(mssql, DateTime.Now, connectBy, "Print jobs", String.Format("Print jobs problem at {0}, times are {1}, repeat No. {2}", connectBy, jobs == "" ? "unknown" : jobs, ts.jobsFailed), ts.jobsFailed, true, false);



                                    if ((ts.nodeParams.doEmail && ts.jobsFailed <= myParams.jobsReports) && (ts.pingFailed == 0) && ((ts.pingFailed == 0) | (!ts.nodeParams.controlPing)))
                                        sendMailFormat(string.Format("Print jobs failed at {0}", connectBy), connectBy, ts.nodeParams.description, DateTime.Now, "Jobs", String.Format("Print jobs problem: {0}", jobs == "" ? "unknown" : jobs), ts.jobsFailed, true);



                                }
                                else
                                {
                                    if (ts.jobsFailed != 0)
                                    {
                                        addEvent(String.Format("Print jobs NOW OK after FAIL at {0}", connectBy), EventLogEntryType.Warning);

                                        if (myParams.writeToSQL)
                                            writeSQL(mssql, DateTime.Now, connectBy, "Print jobs", String.Format("Print jobs NOW OK after FAIL at {0}", connectBy), ts.jobsFailed, false, true);


                                        if (myParams.sendOK)
                                            sendMailFormat(string.Format("Print jobs now OK at {0}", connectBy), connectBy, ts.nodeParams.description, DateTime.Now, "Print jobs", String.Format("Print jobs ok", jobs), ts.jobsFailed, false);
                                    }
                                    else
                                    {
                                        if (myParams.debugMessages)
                                            addEvent(String.Format("Print jobs OK at {0}, times are {1}", connectBy, jobs));
                                    }
                                    ts.jobsFailed = 0;

                                }
                            }
                            else
                            {

                                addEvent(String.Format("Print jobs get info failed at {0}", connectBy), EventLogEntryType.Warning);

                                if (myParams.writeToSQL)
                                    writeSQL(mssql, DateTime.Now, connectBy, "Print jobs", String.Format("Print jobs get info failed at {0}", connectBy), ts.cpjFailed, false, false);

                                if (myParams.reportWhenNoInfo)
                                {

                                    if ((ts.nodeParams.doEmail) && (ts.cpjFailed <= myParams.jobsReports) && ((ts.pingFailed == 0) | (!ts.nodeParams.controlPing)))
                                        sendMailFormat(string.Format("Print jobs get info failed at {0}", connectBy), connectBy, ts.nodeParams.description, DateTime.Now, "Print jobs", String.Format("No information for print jobs"), ts.cpjFailed, true);
                                }

                            }


                        }


                    }

                    catch (Exception exc)
                    {
                        addEvent(String.Format("Timer exception {0} with source {1}, ts = {2}, st = {3}, path = {4}", exc.Message, exc.Source, exc.TargetSite, exc.StackTrace, connectBy), true);
                        if (myParams.writeToSQL)
                            writeSQL(mssql, DateTime.Now, connectBy, "TMS", String.Format("Timer exception {0} with source {1}, ts = {2}, st = {3}, path = {4}", exc.Message, exc.Source, exc.TargetSite, exc.StackTrace, connectBy), 0, false, false);
                    }


                    (sender as System.Timers.Timer).Start();
                    ts.busy = false;
                }
                else
                {
                    addEvent("Timer elapsed at " + DateTime.Now.ToString("HH:mm:ss dd.MM.yyyy") + " for " + connectBy + ", but busy");
                    (sender as System.Timers.Timer).Start();
                }



            }

        }


        public void sendMailDirect(string caption, string message, string attachFile = null, string addr = null)
        {
            try
            {
                MailMessage mail = new MailMessage();
                mail.From = new MailAddress(myParams.mailFrom);

                if (addr == null)
                {

                    foreach (string ma in myParams.contactEmail)
                        mail.To.Add(new MailAddress(ma));
                }
                else
                    mail.To.Add(addr);



                mail.Subject = caption;
                mail.Body = message;
                mail.IsBodyHtml = true;
                mail.BodyEncoding = Encoding.UTF8;

                if (!string.IsNullOrEmpty(attachFile))
                    mail.Attachments.Add(new Attachment(attachFile));
                SmtpClient client = new SmtpClient();
                client.Host = myParams.mailServer;
                client.Port = myParams.mailPort;
                client.EnableSsl = myParams.mailSSL;
                client.Credentials = new NetworkCredential(myParams.mailUser, myParams.mailPassword);
                client.DeliveryMethod = SmtpDeliveryMethod.Network;
                client.Send(mail);
                mail.Dispose();
            }
            catch (Exception exc)
            {

                addEvent(String.Format("Mail exception {0} with source {1}, ts = {2}, st = {3}, path = {4}", exc.Message, exc.Source, exc.TargetSite, exc.StackTrace, addr == null ? "none" : addr), true);
            }
        }



        public void sendMailFormat(string caption, string node, string description, DateTime time, string type, string problem, Int64 tryno, bool isError)
        {
            StreamReader sr = new StreamReader(Assembly.GetExecutingAssembly().GetManifestResourceStream("tms.errorPage.html"));
            var str = sr.ReadToEnd();
            var msg = string.Format(str, node, description, time.ToString("HH:mm:ss dd.MM.yyyy"), type, problem, tryno.ToString("0"), isError ? "lightyellow" : "lightgreen", isError ? "red" : "black");

            sendMailDirect(caption, msg);

        }



        private void checkIncommingEmail()
        {

            try
            {

                if (myParams.debugMessages)
                    addEvent(String.Format("Mail check started "), false);

                Pop3Client client = new Pop3Client();

                client.Connect(myParams.popServer, myParams.popPort, myParams.popSSL);


                client.Authenticate(myParams.popUserName, myParams.popPassWord, AuthenticationMethod.Auto);

                List<OpenPop.Mime.MessageInfo> mis = new List<OpenPop.Mime.MessageInfo>();

                mis = client.GetMessageInfos();


                foreach (OpenPop.Mime.MessageInfo mi in mis)
                {

                    OpenPop.Mime.Header.MessageHeader hdr =
                      client.GetMessageHeaders(mi.Number);

                    if (hdr != null)
                    {

                        if (Regex.IsMatch(hdr.Subject, myParams.popMailSubject))
                        {

                            string body = "";

                            OpenPop.Mime.Message msg = client.GetMessage(mi.Number);

                            if (msg != null)
                            {

                                OpenPop.Mime.MessagePart pt = msg.FindFirstPlainTextVersion();

                                if (pt != null)
                                {
                                    body = pt.GetBodyAsText();

                                }
                                else
                                {
                                    pt = msg.FindFirstHtmlVersion();

                                    if (pt != null)
                                    {

                                        body = pt.GetBodyAsText();
                                        body = Regex.Replace(body, @"<[^>]+>|&nbsp;", "").Trim();
                                        body = Regex.Replace(body, @"\s{2,}", " ");
                                    }


                                }


                                if (Regex.IsMatch(body, myParams.popListAllStatesRegEx))
                                {
                                    if (myParams.debugMessages)
                                        addEvent(String.Format("Found message with good subject and text in body {0}", hdr.Subject), false);

                                    doMailCommand(tmsState.mailCommandTypeEnum.mcListAllStates, hdr.From.Address);

                                    if (myParams.debugMessages)
                                        addEvent(String.Format("Deleting message with number {0}", mi.Number), false);

                                    client.DeleteMessage(mi.Number);


                                }
                                else
                                {
                                    if (myParams.debugMessages)
                                        addEvent(String.Format("Found message with good subject {0}, but bad text", hdr.Subject), false);
                                }


                            }
                            else
                            {

                                if (myParams.debugMessages)
                                    addEvent(String.Format("Can't get message with No. {0}", mi.Number), false);
                            }




                        }
                        else
                        {
                            if (myParams.debugMessages)
                                addEvent(String.Format("Found message with wrong subject {0}", hdr.Subject), false);


                        }
                    }
                    else
                    {
                        if (myParams.debugMessages)
                            addEvent(String.Format("Can't get header of message No. {0}", mi.Number), false);
                    }

                }

                client.Disconnect();
                client.Dispose();

            }
            catch (Exception exc)
            {

                addEvent(String.Format("Mail check exception {0} with source {1}, ts = {2}, st = {3}", exc.Message, exc.Source, exc.TargetSite, exc.StackTrace), true);


            }

        }




        private bool checkJobs(string path, string[] printers, Dictionary<string, Int32> frozen, TimeSpan wmiTimeOut, Int32 jobLimit, out bool withError)
        {
            bool res = true;
            bool error = false;

            try
            {
                var resetEvent = new ManualResetEvent(false);
                ThreadPool.QueueUserWorkItem(
        arg =>
        {
            try
            {
                string providerPath = String.Format(@"\\{0}\root\CIMv2", path);

                string query = string.Format("SELECT Name, TimeSubmitted from Win32_PrintJob");

                ConnectionOptions options = new ConnectionOptions();
                //options.Timeout = new TimeSpan(0, 1, 0);
                ManagementScope scope = new ManagementScope(providerPath, options);
                ManagementObjectSearcher searcher = new ManagementObjectSearcher(scope, new ObjectQuery(query));
                searcher.Options.Timeout = wmiTimeOut;
                searcher.Options.ReturnImmediately = false;
                ManagementObjectCollection coll = searcher.Get();
                foreach (ManagementObject printer in coll)
                {
                    string pName = printer.Properties["Name"].Value as string;
                    foreach (string rm in printers)
                    {
                        if (Regex.IsMatch(pName, rm))
                        {
                            Int32 mins = Convert.ToInt32((DateTime.Now - ManagementDateTimeConverter.ToDateTime(printer.Properties["TimeSubmitted"].Value as string)).TotalMinutes);

                            if (mins > jobLimit)
                            {
                                frozen.Add(pName, mins);
                                res = false;
                            }

                            break;
                        }
                    }
                }
            }

            catch (Exception exc)
            {
                addEvent(String.Format("Check print jobs exception {0} with source {1}, ts = {2}, st = {3}, path = {4}", exc.Message, exc.Source, exc.TargetSite, exc.StackTrace, path), true);
                res = false;
                error = true;
            };

            resetEvent.Set();

        });
                resetEvent.WaitOne();
            }
            catch (Exception exc)
            {
                addEvent(String.Format("Check print jobs exception {0} with source {1}, ts = {2}, st = {3}, path = {4}", exc.Message, exc.Source, exc.TargetSite, exc.StackTrace, path), true);
                res = false;
                error = true;
            }

            withError = error;
            return res;

        }


        private bool checkJobsProxy(string path, string[] printers, Dictionary<string, Int32> frozen, TimeSpan wmiTimeOut, TimeSpan waitTime, Int32 jobLimit, out bool withError)
        {
            bool error = false;
            try
            {
                Stopwatch sw = new Stopwatch();
                Int32 csInterval = Convert.ToInt32(waitTime.TotalMilliseconds);
                bool csFlag = false;
                Int32 tryNo = 0;

                sw.Start();

                while ((sw.ElapsedMilliseconds <= csInterval) && (!csFlag))
                {
                    tryNo++;
                    if (myParams.debugMessages)
                        addEvent(string.Format("Trying to check print jobs on path {0}, try No. {1}", path, tryNo));

                    frozen.Clear();
                    csFlag = checkJobs(path, printers, frozen, wmiTimeOut, jobLimit, out error);
                }
                sw.Stop();
                withError = error;
                return csFlag;
            }

            catch (Exception exc)
            {
                addEvent(String.Format("Check print jobs proxy exception {0} with source {1}, ts = {2}, st = {3}, path = {4}", exc.Message, exc.Source, exc.TargetSite, exc.StackTrace, path), true);
                withError = true;
                return false;
            }
        }




    }
}
